# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from abc import ABC

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.bladed_model import BladedModel
class CampbellDiagramModeOfOperation_CampbellDiagramModeOfOperationTypeEnum(str, Enum):
    IDLING = "Idling"
    PARKED = "Parked"
    POWER_PRODUCTION = "PowerProduction"
    INSERT = "Insert"

from .schema_helper import SchemaHelper
from .models_impl import *


class CampbellDiagramModeOfOperation(BladedModel, ABC):
    r"""
    The common properties of the operating mode of a Campbell diagram calculation.
    
    Not supported yet.
    
    Attributes
    ----------
    CampbellDiagramModeOfOperationType : CampbellDiagramModeOfOperation_CampbellDiagramModeOfOperationTypeEnum, Not supported yet
        Defines the specific type of model in use.
    
    Notes
    -----
    This model is not supported yet by the Bladed calculation engine.
    This class is an abstraction, with the following concrete implementations:
        - CampbellIdling
        - CampbellDiagramModeOfOperationInsert
        - CampbellParked
        - CampbellPowerProduction
    
    """
    CampbellDiagramModeOfOperationType: CampbellDiagramModeOfOperation_CampbellDiagramModeOfOperationTypeEnum = Field(alias="CampbellDiagramModeOfOperationType", default=None) # Not supported yet

    _type_info = TypeInfo(
        set([]),
        set([]),
        set([]),
        None).merge(BladedModel._type_info)


CampbellDiagramModeOfOperation.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from abc import ABC

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.bladed_model import BladedModel
from dnv_bladed_models.sequential_timestep_adaptation import SequentialTimestepAdaptation
class AerodynamicModel_AerodynamicModelTypeEnum(str, Enum):
    BLADE_ELEMENT_MOMENTUM = "BladeElementMomentum"
    VORTEX_LINE = "VortexLine"
    INSERT = "Insert"

from .schema_helper import SchemaHelper
from .models_impl import *


class AerodynamicModel(BladedModel, ABC):
    r"""
    Common properties for all aerodynamics models.
    
    Attributes
    ----------
    AerodynamicModelType : AerodynamicModel_AerodynamicModelTypeEnum
        Defines the specific type of model in use.
    
    SequentialTimestepAdaptation : SequentialTimestepAdaptation, Not supported yet
    
    Notes
    -----
    
    This class is an abstraction, with the following concrete implementations:
        - BladeElementMomentum
        - AerodynamicModelInsert
        - VortexLine
    
    """
    AerodynamicModelType: AerodynamicModel_AerodynamicModelTypeEnum = Field(alias="AerodynamicModelType", default=None)
    SequentialTimestepAdaptation: SequentialTimestepAdaptation = Field(alias="SequentialTimestepAdaptation", default=None) # Not supported yet

    _type_info = TypeInfo(
        set([]),
        set([]),
        set([]),
        None).merge(BladedModel._type_info)


AerodynamicModel.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.bladed_model import BladedModel

from .schema_helper import SchemaHelper
from .models_impl import *


class TowerNodeLoads(BladedModel):
    r"""
    The moments and forces to output for any node location where the loads are requested.
    
    Not supported yet.
    
    Attributes
    ----------
    Mx : bool, default=False, Not supported yet
        If true, the moment about the x-axis will be output.
    
    My : bool, default=False, Not supported yet
        If true, the moment about the y-axis will be output.
    
    Mz : bool, default=False, Not supported yet
        If true, the moment about the z-axis will be output.
    
    Mxy : bool, Not supported yet
        If true, the maximum moment about any axis that lies in the XY plane will be output.
    
    Fx : bool, default=False, Not supported yet
        If true, the force in the X direction will be output.
    
    Fy : bool, default=False, Not supported yet
        If true, the force in the Y direction will be output.
    
    Fz : bool, default=False, Not supported yet
        If true, the force in the Z direction will be output.
    
    Fxy : bool, Not supported yet
        If true, the maximum force along any axis that lies in the XY plane will be output.
    
    Notes
    -----
    This model is not supported yet by the Bladed calculation engine.
    """
    Mx: bool = Field(alias="Mx", default=None) # Not supported yet
    My: bool = Field(alias="My", default=None) # Not supported yet
    Mz: bool = Field(alias="Mz", default=None) # Not supported yet
    Mxy: bool = Field(alias="Mxy", default=None) # Not supported yet
    Fx: bool = Field(alias="Fx", default=None) # Not supported yet
    Fy: bool = Field(alias="Fy", default=None) # Not supported yet
    Fz: bool = Field(alias="Fz", default=None) # Not supported yet
    Fxy: bool = Field(alias="Fxy", default=None) # Not supported yet

    _relative_schema_path = 'Components/Tower/TowerOutputGroupLibrary/TowerOutputGroup/TowerNodeLoads/TowerNodeLoads.json'
    _type_info = TypeInfo(
        set([]),
        set([]),
        set([]),
        None).merge(BladedModel._type_info)


    class Config:
        extra = Extra.forbid
        validate_assignment = True
        allow_population_by_field_name = True
        pass

    def _entity(self) -> bool:
        return True


TowerNodeLoads.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from abc import ABC

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.current import Current
from dnv_bladed_models.current_direction_time_history import CurrentDirectionTimeHistory
from dnv_bladed_models.current_direction_variation import CurrentDirectionVariation
from dnv_bladed_models.current_direction_variation_insert import CurrentDirectionVariationInsert
from dnv_bladed_models.preset_current_direction_transient import PresetCurrentDirectionTransient

from .schema_helper import SchemaHelper
from .models_impl import *

TCurrentDirectionVariationOptions = TypeVar('TCurrentDirectionVariationOptions', CurrentDirectionVariationInsert, PresetCurrentDirectionTransient, CurrentDirectionTimeHistory, CurrentDirectionVariation, )

class TimeDomainCurrent(Current, ABC):
    r"""
    The definition of a current field that varies throughout a time domain simulation.
    
    Not supported yet.
    
    Attributes
    ----------
    ReferenceHeight : float, Not supported yet
        The reference height for the current field, above which the free-field current conditions take over.  If this is omitted, the hub height will be used, and if there is more than one the *highest* hub height.
    
    Inclination : float, default=0, Not supported yet
        The inclination of the flow relative to the horizontal plane.
    
    Direction : float, Not supported yet
        The (constant) direction of the current relative to the global x-axis.
    
    DirectionVariation : Union[CurrentDirectionVariationInsert, PresetCurrentDirectionTransient, CurrentDirectionTimeHistory], Not supported yet
    
    Notes
    -----
    This model is not supported yet by the Bladed calculation engine.
    """
    ReferenceHeight: float = Field(alias="ReferenceHeight", default=None) # Not supported yet
    Inclination: float = Field(alias="Inclination", default=None) # Not supported yet
    Direction: float = Field(alias="Direction", default=None) # Not supported yet
    DirectionVariation: Union[CurrentDirectionVariationInsert, PresetCurrentDirectionTransient, CurrentDirectionTimeHistory] = Field(alias="DirectionVariation", default=None, discriminator='DirectionVariationType') # Not supported yet

    _type_info = TypeInfo(
        set([('DirectionVariation', 'DirectionVariationType'),]),
        set([]),
        set([]),
        None).merge(Current._type_info)


    @property
    def DirectionVariation_as_PresetCurrentDirectionTransient(self) -> PresetCurrentDirectionTransient:
        """
        Retrieves the value of DirectionVariation guaranteeing it is a PresetCurrentDirectionTransient; if it is not, an error is raised.

        Useful when using type-checking development tools, as it provides a concise way to retrieve a value and assign a variable with the correct type.

        Returns
        -------
        PresetCurrentDirectionTransient
            A model object, guaranteed to be a PresetCurrentDirectionTransient.

        Raises
        ------
        TypeError
            If the value is not a PresetCurrentDirectionTransient.
        """
        return self.DirectionVariation_as(PresetCurrentDirectionTransient)


    @property
    def DirectionVariation_as_CurrentDirectionTimeHistory(self) -> CurrentDirectionTimeHistory:
        """
        Retrieves the value of DirectionVariation guaranteeing it is a CurrentDirectionTimeHistory; if it is not, an error is raised.

        Useful when using type-checking development tools, as it provides a concise way to retrieve a value and assign a variable with the correct type.

        Returns
        -------
        CurrentDirectionTimeHistory
            A model object, guaranteed to be a CurrentDirectionTimeHistory.

        Raises
        ------
        TypeError
            If the value is not a CurrentDirectionTimeHistory.
        """
        return self.DirectionVariation_as(CurrentDirectionTimeHistory)


    @property
    def DirectionVariation_as_inline(self) -> Union[PresetCurrentDirectionTransient, CurrentDirectionTimeHistory]:
        """
        Retrieves the value of DirectionVariation as a model object; if the value is specified with a '$insert', an error is raised.

        Returns
        -------
        Union[PresetCurrentDirectionTransient, CurrentDirectionTimeHistory]
            A model object at the specified index, guaranteed to not be a '$insert'.

        Raises
        ------
        TypeError
            If the value is not one of the concrete types of CurrentDirectionVariation; i.e. it is specified with a '$insert'.
        """
        if isinstance(self.DirectionVariation, CurrentDirectionVariationInsert) or self.DirectionVariation.is_insert:
            raise TypeError(f"Expected DirectionVariation value to be an in-line object, but it is currently in the '$insert' state.")
        return self.DirectionVariation


    def DirectionVariation_as(self, cls: Type[TCurrentDirectionVariationOptions])-> TCurrentDirectionVariationOptions:
        """
        Retrieves the value of DirectionVariation, ensuring it is of the specified type.
        The specified type must be one of the valid concrete types of CurrentDirectionVariation, or the '$insert' type
        (in the case the model is defined to be inserted from an external resource).

        Useful when using type-checking development tools, as it provides a concise way to access a value with the correct type.

        Parameters
        ----------
        cls: Type[Union[CurrentDirectionVariationInsert, PresetCurrentDirectionTransient, CurrentDirectionTimeHistory]]
            One of the valid concrete types of CurrentDirectionVariation, or the '$insert' type
            (in the case the model is defined to be inserted from an external resource).

        Returns
        -------
        TCurrentDirectionVariationOptions
            A model object of the specified type.

        Raises
        ------
        TypeError
            If the value is not of the specified type.

        Examples
        --------
        Get a reference to the value when it is one of the types of CurrentDirectionVariation:
        >>> val_obj = model_obj.DirectionVariation_as(models.PresetCurrentDirectionTransient)
        >>> val_obj = model_obj.DirectionVariation_as(models.CurrentDirectionTimeHistory)

        Get a reference to the value, when it was specified with a '$insert' and read in from a file:
        >>> insert_obj = model_obj.DirectionVariation_as(models.CurrentDirectionVariationInsert)
        """
        if not isinstance(self.DirectionVariation, cls):
            raise TypeError(f"Expected DirectionVariation of type '{cls.__name__}' but was type '{type(self.DirectionVariation).__name__}'")
        return self.DirectionVariation


TimeDomainCurrent.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, TypeVar, Union
from abc import ABC, abstractmethod
from pydantic import BaseModel, ValidationError, root_validator, PrivateAttr
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
TPydanticModel = TypeVar('TPydanticModel', bound='BaseModel')
TEntity = TypeVar('TEntity', bound='BladedEntity')

@dataclass(frozen=True)
class TypeInfo:
    discriminated_props: set[Tuple[str, str]]
    discriminated_arrays: set[Tuple[str, str]]
    containers: set[str]
    discriminator: Optional[str] = None

    def merge(self, other: 'TypeInfo') -> 'TypeInfo':
        return TypeInfo(
            self.discriminated_props.union(other.discriminated_props),
            self.discriminated_arrays.union(other.discriminated_arrays),
            self.containers.union(other.containers),
            other.discriminator if self.discriminator is None else self.discriminator)


class BladedEntity(BaseModel, ABC):

    _incoming_fields: set[str] = PrivateAttr(default_factory=lambda: set())
    _type_info: TypeInfo = TypeInfo(set(), set(), set())
   
    @abstractmethod
    def _entity(self) -> bool:
        pass


    @root_validator(pre=True)
    def _remove_underscore_fields(cls, values: Dict[str, Any]):
        remove_underscore_fields(values)
        return values


    @classmethod
    def _model_factory(cls: Type['TPydanticModel'], obj: dict[str, Any]) -> TPydanticModel:
        return super().parse_obj(obj)


    @classmethod
    def parse_obj(cls: Type['TEntity'], obj: dict[str, Any]) -> 'TEntity':
        """
        Create a new model instance from a dictionary of properties.
        """
        obj_as_dict = prepare_model_dict(cls, obj, cls._type_info.discriminated_props, cls._type_info.discriminated_arrays)
        model = cls._model_factory(obj_as_dict)
        model._incoming_fields.update(obj_as_dict.keys())
        return model


    @classmethod
    def __get_validators__(cls):
        yield cls._validator


    @classmethod
    def _validator(cls, data):
        if isinstance(data, cls):
            return data
        return cls.parse_obj(data)


    @property
    def is_insert(self) -> bool:
        """
        Returns true if the model is to be loaded from an external resource by the Bladed application; i.e. the 'insert' field is set with a resource location.
        """
        if 'insert' in self.__fields_set__:
            insert_val = self.__dict__['insert']
            return insert_val is not None and insert_val != ''
        return False

    
    def _find_unused_containers(self) -> Set[str]:
        unused_containers: Set[str] = set()
        for container in self._type_info.containers:
            if container in self.__dict__:
                container_obj = self.__dict__[container]
                if container_obj is not None and len(container_obj) == 0 and container not in self._incoming_fields:
                    if isinstance(container_obj, BladedEntity):
                        if container_obj._is_unused():
                            unused_containers.add(container)
                    else:
                        unused_containers.add(container)
        return unused_containers


    def _is_unused(self) -> bool:
        """
        Returns true if the model has no user-supplied values. This indicates the object should not be rendered in the final JSON output.
        """
        candidate_fields = [x[1] for x in self.__dict__.items() if x[0] != self._type_info.discriminator]
        return (len(candidate_fields) == 0 or all(map(lambda f: f is None, candidate_fields)))


    def _iter( # type: ignore
        self,
        **kwargs: Any
    ):
        if self.is_insert:
            kwargs['exclude'] = None
            kwargs['include'] = set(['insert'])
        else:
            exclude: Optional[Set[str]] = kwargs.get('exclude', set())
            if exclude is None:
                exclude = self._find_unused_containers()
            else:
                exclude.update(self._find_unused_containers())
            kwargs['exclude'] = exclude
        return super()._iter(**kwargs)


def remove_underscore_fields(values: Dict[str, Any]):
    to_remove: Set[str] = set()
    for child_name, child in values.items():
        if child_name.startswith('_'):
            to_remove.add(child_name)
        elif isinstance(child, dict):
            remove_underscore_fields(child)
        elif isinstance(child, list):
            for item in child:
                if isinstance(item, dict):
                    remove_underscore_fields(item)
    for x in to_remove:
        del values[x]


def prepare_dict_for_discriminated_insert(cls, field_name: str, field_obj: dict, discriminator_prop: str):
    if isinstance(field_obj, dict) and '$insert' in field_obj:
        if discriminator_prop in field_obj and field_obj[discriminator_prop] is not None and field_obj['$insert'] is not None:
            exc = ValueError(f"Cannot set both {discriminator_prop} and $insert fields.")
            raise ValidationError([ErrorWrapper(exc, loc=field_name)], cls)
        field_obj[discriminator_prop] = 'Insert'


def prepare_list_for_discriminated_insert(cls, field_name: str, field_obj: dict, discriminator_prop: str):
    if isinstance(field_obj, list):
        i = 0
        for item in field_obj:
            if isinstance(item, dict) and '$insert' in item:
                if discriminator_prop in item and item[discriminator_prop] is not None and item['$insert'] is not None:
                    exc = ValueError(f"Cannot set both {discriminator_prop} and $insert fields.")
                    raise ValidationError([ErrorWrapper(exc, loc=f"{field_name}[{i}]")], cls)
                item[discriminator_prop] = 'Insert'
            i += 1
    

def ensure_dict_for_parse(cls, data) -> dict:
    obj = cls._enforce_dict_if_root(data)
    if not isinstance(obj, dict):
        try:
            obj = dict(obj)
        except (TypeError, ValueError) as e:
            exc = TypeError(f'{cls.__name__} expected dict not {obj.__class__.__name__}')
            raise ValidationError([ErrorWrapper(exc, loc=ROOT_KEY)], cls) from e
    return obj


TParsedObj = TypeVar('TParsedObj', bound=BaseModel)
def prepare_model_dict(cls : Type[TParsedObj], obj: Any, discriminated_props: set[Tuple[str, str]], discriminated_arrays: set[Tuple[str, str]]) -> dict:
    obj_as_dict = ensure_dict_for_parse(cls, obj)
    if any(discriminated_props) or any(discriminated_arrays):
        for field_name, discriminator in discriminated_props:
            if field_name in obj_as_dict:
                prepare_dict_for_discriminated_insert(cls, field_name, obj_as_dict[field_name], discriminator)
        for field_name, discriminator in discriminated_arrays:
            if field_name in obj_as_dict:
                prepare_list_for_discriminated_insert(cls, field_name, obj_as_dict[field_name], discriminator)
    return obj_as_dict


TRawContainer = TypeVar('TRawContainer')
def custom_entries_parser(cls, obj, valid_raw_type: Type[TRawContainer], entry_model: Type, prepare_data: Optional[Callable[[str, TRawContainer], None]] = None):
    data_dict = ensure_dict_for_parse(cls, obj)

    ctor_data = {}
    entry_data = {}
    field_keys = set(val.alias or val.name for val in cls.__fields__.values())
    for key, val in data_dict.items():
        if not key in field_keys and not key.startswith('_') and isinstance(val, valid_raw_type):
            if prepare_data is not None:
                prepare_data(key, val)
            entry_data[key] = val
        else:
            ctor_data[key] = val

    instance = cls(**ctor_data)
    try:
        container = entry_model.parse_obj({ 'entries' : entry_data })
        for key, val in container.entries.items():
            setattr(instance, key, val)
    except ValidationError as e:
        raise ValidationError([ErrorWrapper(e, loc=cls.__name__)], cls)
           
    return instance





# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.circular_lidar_scan import CircularLidarScan
from dnv_bladed_models.component import Component
from dnv_bladed_models.controller_lidar_settings import ControllerLidarSettings
from dnv_bladed_models.lidar_beam import LidarBeam
from dnv_bladed_models.lidar_controller_scan import LidarControllerScan
from dnv_bladed_models.lidar_focal_distance_control import LidarFocalDistanceControl
from dnv_bladed_models.lidar_focal_distance_control_insert import LidarFocalDistanceControlInsert
from dnv_bladed_models.lidar_scanning_pattern import LidarScanningPattern
from dnv_bladed_models.lidar_scanning_pattern_insert import LidarScanningPatternInsert
from dnv_bladed_models.look_up_table_element import LookUpTableElement
from dnv_bladed_models.multiple_lidar_focal_distances import MultipleLidarFocalDistances
from dnv_bladed_models.rosette_lidar_scan import RosetteLidarScan
from dnv_bladed_models.single_lidar_focal_distance import SingleLidarFocalDistance

from .schema_helper import SchemaHelper
from .models_impl import *

TLidarScanningPatternOptions = TypeVar('TLidarScanningPatternOptions', CircularLidarScan, LidarControllerScan, LidarScanningPatternInsert, RosetteLidarScan, LidarScanningPattern, )
TLidarFocalDistanceControlOptions = TypeVar('TLidarFocalDistanceControlOptions', ControllerLidarSettings, LidarFocalDistanceControlInsert, MultipleLidarFocalDistances, SingleLidarFocalDistance, LidarFocalDistanceControl, )

class Lidar(Component):
    r"""
    A Lidar sensor mounted on the structure.
    
    Not supported yet.
    
    Attributes
    ----------
    ComponentType : Literal['Lidar'], default='Lidar', Not supported yet
        Defines the specific type of Component model in use.  For a `Lidar` object, this must always be set to a value of `Lidar`.
    
    LensArea : float, Not supported yet
        The area of the lens.
    
    LaserWavelength : float, Not supported yet
        The wavelength of the laser.
    
    WeightingFunction : List[LookUpTableElement], Not supported yet
        The relationship between the distance from the focal point and the weighting to put on the sample.  Every Lidar sampling point has a finite width, where velocities are collected to either side of the nominal focal distance.  This relationship is used to put more weight on those samples closest to the nominal focal point.
    
    LidarBeams : List[LidarBeam], Not supported yet
        The definition of the lidar beams.
    
    ScanningPattern : Union[CircularLidarScan, LidarControllerScan, LidarScanningPatternInsert, RosetteLidarScan], Not supported yet
    
    FocalDistanceControl : Union[ControllerLidarSettings, LidarFocalDistanceControlInsert, MultipleLidarFocalDistances, SingleLidarFocalDistance], Not supported yet
    
    Notes
    -----
    This model is not supported yet by the Bladed calculation engine.
    """
    ComponentType: Literal['Lidar'] = Field(alias="ComponentType", default='Lidar', allow_mutation=False, const=True) # Not supported yet # type: ignore
    LensArea: float = Field(alias="LensArea", default=None) # Not supported yet
    LaserWavelength: float = Field(alias="LaserWavelength", default=None) # Not supported yet
    WeightingFunction: List[LookUpTableElement] = Field(alias="WeightingFunction", default=list()) # Not supported yet
    LidarBeams: List[LidarBeam] = Field(alias="LidarBeams", default=list()) # Not supported yet
    ScanningPattern: Union[CircularLidarScan, LidarControllerScan, LidarScanningPatternInsert, RosetteLidarScan] = Field(alias="ScanningPattern", default=None, discriminator='ScanningPatternType') # Not supported yet
    FocalDistanceControl: Union[ControllerLidarSettings, LidarFocalDistanceControlInsert, MultipleLidarFocalDistances, SingleLidarFocalDistance] = Field(alias="FocalDistanceControl", default=None, discriminator='FocalDistanceControlType') # Not supported yet

    _relative_schema_path = 'Components/Lidar/Lidar.json'
    _type_info = TypeInfo(
        set([('ScanningPattern', 'ScanningPatternType'),('FocalDistanceControl', 'FocalDistanceControlType'),]),
        set([]),
        set(['WeightingFunction','LidarBeams',]),
        'ComponentType').merge(Component._type_info)


    class Config:
        extra = Extra.forbid
        validate_assignment = True
        allow_population_by_field_name = True
        pass

    @property
    def ScanningPattern_as_CircularLidarScan(self) -> CircularLidarScan:
        """
        Retrieves the value of ScanningPattern guaranteeing it is a CircularLidarScan; if it is not, an error is raised.

        Useful when using type-checking development tools, as it provides a concise way to retrieve a value and assign a variable with the correct type.

        Returns
        -------
        CircularLidarScan
            A model object, guaranteed to be a CircularLidarScan.

        Raises
        ------
        TypeError
            If the value is not a CircularLidarScan.
        """
        return self.ScanningPattern_as(CircularLidarScan)


    @property
    def ScanningPattern_as_LidarControllerScan(self) -> LidarControllerScan:
        """
        Retrieves the value of ScanningPattern guaranteeing it is a LidarControllerScan; if it is not, an error is raised.

        Useful when using type-checking development tools, as it provides a concise way to retrieve a value and assign a variable with the correct type.

        Returns
        -------
        LidarControllerScan
            A model object, guaranteed to be a LidarControllerScan.

        Raises
        ------
        TypeError
            If the value is not a LidarControllerScan.
        """
        return self.ScanningPattern_as(LidarControllerScan)


    @property
    def ScanningPattern_as_RosetteLidarScan(self) -> RosetteLidarScan:
        """
        Retrieves the value of ScanningPattern guaranteeing it is a RosetteLidarScan; if it is not, an error is raised.

        Useful when using type-checking development tools, as it provides a concise way to retrieve a value and assign a variable with the correct type.

        Returns
        -------
        RosetteLidarScan
            A model object, guaranteed to be a RosetteLidarScan.

        Raises
        ------
        TypeError
            If the value is not a RosetteLidarScan.
        """
        return self.ScanningPattern_as(RosetteLidarScan)


    @property
    def ScanningPattern_as_inline(self) -> Union[CircularLidarScan, LidarControllerScan, RosetteLidarScan]:
        """
        Retrieves the value of ScanningPattern as a model object; if the value is specified with a '$insert', an error is raised.

        Returns
        -------
        Union[CircularLidarScan, LidarControllerScan, RosetteLidarScan]
            A model object at the specified index, guaranteed to not be a '$insert'.

        Raises
        ------
        TypeError
            If the value is not one of the concrete types of LidarScanningPattern; i.e. it is specified with a '$insert'.
        """
        if isinstance(self.ScanningPattern, LidarScanningPatternInsert) or self.ScanningPattern.is_insert:
            raise TypeError(f"Expected ScanningPattern value to be an in-line object, but it is currently in the '$insert' state.")
        return self.ScanningPattern


    def ScanningPattern_as(self, cls: Type[TLidarScanningPatternOptions])-> TLidarScanningPatternOptions:
        """
        Retrieves the value of ScanningPattern, ensuring it is of the specified type.
        The specified type must be one of the valid concrete types of LidarScanningPattern, or the '$insert' type
        (in the case the model is defined to be inserted from an external resource).

        Useful when using type-checking development tools, as it provides a concise way to access a value with the correct type.

        Parameters
        ----------
        cls: Type[Union[CircularLidarScan, LidarControllerScan, LidarScanningPatternInsert, RosetteLidarScan]]
            One of the valid concrete types of LidarScanningPattern, or the '$insert' type
            (in the case the model is defined to be inserted from an external resource).

        Returns
        -------
        TLidarScanningPatternOptions
            A model object of the specified type.

        Raises
        ------
        TypeError
            If the value is not of the specified type.

        Examples
        --------
        Get a reference to the value when it is one of the types of LidarScanningPattern:
        >>> val_obj = model_obj.ScanningPattern_as(models.CircularLidarScan)
        >>> val_obj = model_obj.ScanningPattern_as(models.LidarControllerScan)
        >>> val_obj = model_obj.ScanningPattern_as(models.RosetteLidarScan)

        Get a reference to the value, when it was specified with a '$insert' and read in from a file:
        >>> insert_obj = model_obj.ScanningPattern_as(models.LidarScanningPatternInsert)
        """
        if not isinstance(self.ScanningPattern, cls):
            raise TypeError(f"Expected ScanningPattern of type '{cls.__name__}' but was type '{type(self.ScanningPattern).__name__}'")
        return self.ScanningPattern


    @property
    def FocalDistanceControl_as_ControllerLidarSettings(self) -> ControllerLidarSettings:
        """
        Retrieves the value of FocalDistanceControl guaranteeing it is a ControllerLidarSettings; if it is not, an error is raised.

        Useful when using type-checking development tools, as it provides a concise way to retrieve a value and assign a variable with the correct type.

        Returns
        -------
        ControllerLidarSettings
            A model object, guaranteed to be a ControllerLidarSettings.

        Raises
        ------
        TypeError
            If the value is not a ControllerLidarSettings.
        """
        return self.FocalDistanceControl_as(ControllerLidarSettings)


    @property
    def FocalDistanceControl_as_MultipleLidarFocalDistances(self) -> MultipleLidarFocalDistances:
        """
        Retrieves the value of FocalDistanceControl guaranteeing it is a MultipleLidarFocalDistances; if it is not, an error is raised.

        Useful when using type-checking development tools, as it provides a concise way to retrieve a value and assign a variable with the correct type.

        Returns
        -------
        MultipleLidarFocalDistances
            A model object, guaranteed to be a MultipleLidarFocalDistances.

        Raises
        ------
        TypeError
            If the value is not a MultipleLidarFocalDistances.
        """
        return self.FocalDistanceControl_as(MultipleLidarFocalDistances)


    @property
    def FocalDistanceControl_as_SingleLidarFocalDistance(self) -> SingleLidarFocalDistance:
        """
        Retrieves the value of FocalDistanceControl guaranteeing it is a SingleLidarFocalDistance; if it is not, an error is raised.

        Useful when using type-checking development tools, as it provides a concise way to retrieve a value and assign a variable with the correct type.

        Returns
        -------
        SingleLidarFocalDistance
            A model object, guaranteed to be a SingleLidarFocalDistance.

        Raises
        ------
        TypeError
            If the value is not a SingleLidarFocalDistance.
        """
        return self.FocalDistanceControl_as(SingleLidarFocalDistance)


    @property
    def FocalDistanceControl_as_inline(self) -> Union[ControllerLidarSettings, MultipleLidarFocalDistances, SingleLidarFocalDistance]:
        """
        Retrieves the value of FocalDistanceControl as a model object; if the value is specified with a '$insert', an error is raised.

        Returns
        -------
        Union[ControllerLidarSettings, MultipleLidarFocalDistances, SingleLidarFocalDistance]
            A model object at the specified index, guaranteed to not be a '$insert'.

        Raises
        ------
        TypeError
            If the value is not one of the concrete types of LidarFocalDistanceControl; i.e. it is specified with a '$insert'.
        """
        if isinstance(self.FocalDistanceControl, LidarFocalDistanceControlInsert) or self.FocalDistanceControl.is_insert:
            raise TypeError(f"Expected FocalDistanceControl value to be an in-line object, but it is currently in the '$insert' state.")
        return self.FocalDistanceControl


    def FocalDistanceControl_as(self, cls: Type[TLidarFocalDistanceControlOptions])-> TLidarFocalDistanceControlOptions:
        """
        Retrieves the value of FocalDistanceControl, ensuring it is of the specified type.
        The specified type must be one of the valid concrete types of LidarFocalDistanceControl, or the '$insert' type
        (in the case the model is defined to be inserted from an external resource).

        Useful when using type-checking development tools, as it provides a concise way to access a value with the correct type.

        Parameters
        ----------
        cls: Type[Union[ControllerLidarSettings, LidarFocalDistanceControlInsert, MultipleLidarFocalDistances, SingleLidarFocalDistance]]
            One of the valid concrete types of LidarFocalDistanceControl, or the '$insert' type
            (in the case the model is defined to be inserted from an external resource).

        Returns
        -------
        TLidarFocalDistanceControlOptions
            A model object of the specified type.

        Raises
        ------
        TypeError
            If the value is not of the specified type.

        Examples
        --------
        Get a reference to the value when it is one of the types of LidarFocalDistanceControl:
        >>> val_obj = model_obj.FocalDistanceControl_as(models.ControllerLidarSettings)
        >>> val_obj = model_obj.FocalDistanceControl_as(models.MultipleLidarFocalDistances)
        >>> val_obj = model_obj.FocalDistanceControl_as(models.SingleLidarFocalDistance)

        Get a reference to the value, when it was specified with a '$insert' and read in from a file:
        >>> insert_obj = model_obj.FocalDistanceControl_as(models.LidarFocalDistanceControlInsert)
        """
        if not isinstance(self.FocalDistanceControl, cls):
            raise TypeError(f"Expected FocalDistanceControl of type '{cls.__name__}' but was type '{type(self.FocalDistanceControl).__name__}'")
        return self.FocalDistanceControl


    def _entity(self) -> bool:
        return True


Lidar.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from abc import ABC

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.bladed_model import BladedModel
class InitialCondition_InitialConditionTypeEnum(str, Enum):
    GL2010_ICING = "GL2010Icing"
    IEC4_ICING = "IEC4Icing"
    INITIAL_AZIMUTH_POSITION = "InitialAzimuthPosition"
    INITIAL_FLOATING_POSITION = "InitialFloatingPosition"
    INITIAL_PITCH_POSITION = "InitialPitchPosition"
    INITIAL_ROTOR_SPEED = "InitialRotorSpeed"
    INITIAL_YAW_ANGLE = "InitialYawAngle"
    ROTOR_IDLING = "RotorIdling"
    ROTOR_IN_POWER_PRODUCTION = "RotorInPowerProduction"
    ROTOR_PARKED = "RotorParked"
    INSERT = "Insert"

from .schema_helper import SchemaHelper
from .models_impl import *


class InitialCondition(BladedModel, ABC):
    r"""
    A condition of the turbine or a component at the beginning of the simulation.  This may change during the simulation.
    
    Attributes
    ----------
    InitialConditionType : InitialCondition_InitialConditionTypeEnum
        Defines the specific type of model in use.
    
    Notes
    -----
    
    This class is an abstraction, with the following concrete implementations:
        - GL2010Icing
        - IEC4Icing
        - InitialAzimuthPosition
        - InitialFloatingPosition
        - InitialPitchPosition
        - InitialRotorSpeed
        - InitialYawAngle
        - InitialConditionInsert
        - RotorIdlingControlState
        - RotorInPowerProduction
        - RotorParkedControlState
    
    """
    InitialConditionType: InitialCondition_InitialConditionTypeEnum = Field(alias="InitialConditionType", default=None)

    _type_info = TypeInfo(
        set([]),
        set([]),
        set([]),
        None).merge(BladedModel._type_info)


InitialCondition.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from abc import ABC

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.component import Component
from dnv_bladed_models.modal_structural_modelling import ModalStructuralModelling
from dnv_bladed_models.rigid_structural_modelling import RigidStructuralModelling
from dnv_bladed_models.structural_modelling import StructuralModelling
from dnv_bladed_models.structural_modelling_insert import StructuralModellingInsert

from .schema_helper import SchemaHelper
from .models_impl import *

TStructuralModellingOptions = TypeVar('TStructuralModellingOptions', StructuralModellingInsert, ModalStructuralModelling, RigidStructuralModelling, StructuralModelling, )

class FlexBody(Component, ABC):
    r"""
    The common properties of components modelled using multibody flex bodies.
    
    Attributes
    ----------
    Modelling : Union[StructuralModellingInsert, ModalStructuralModelling, RigidStructuralModelling]
    
    Notes
    -----
    
    """
    Modelling: Union[StructuralModellingInsert, ModalStructuralModelling, RigidStructuralModelling] = Field(alias="Modelling", default=None, discriminator='StructuralModellingType')

    _type_info = TypeInfo(
        set([('Modelling', 'StructuralModellingType'),]),
        set([]),
        set([]),
        None).merge(Component._type_info)


    @property
    def Modelling_as_ModalStructuralModelling(self) -> ModalStructuralModelling:
        """
        Retrieves the value of Modelling guaranteeing it is a ModalStructuralModelling; if it is not, an error is raised.

        Useful when using type-checking development tools, as it provides a concise way to retrieve a value and assign a variable with the correct type.

        Returns
        -------
        ModalStructuralModelling
            A model object, guaranteed to be a ModalStructuralModelling.

        Raises
        ------
        TypeError
            If the value is not a ModalStructuralModelling.
        """
        return self.Modelling_as(ModalStructuralModelling)


    @property
    def Modelling_as_RigidStructuralModelling(self) -> RigidStructuralModelling:
        """
        Retrieves the value of Modelling guaranteeing it is a RigidStructuralModelling; if it is not, an error is raised.

        Useful when using type-checking development tools, as it provides a concise way to retrieve a value and assign a variable with the correct type.

        Returns
        -------
        RigidStructuralModelling
            A model object, guaranteed to be a RigidStructuralModelling.

        Raises
        ------
        TypeError
            If the value is not a RigidStructuralModelling.
        """
        return self.Modelling_as(RigidStructuralModelling)


    @property
    def Modelling_as_inline(self) -> Union[ModalStructuralModelling, RigidStructuralModelling]:
        """
        Retrieves the value of Modelling as a model object; if the value is specified with a '$insert', an error is raised.

        Returns
        -------
        Union[ModalStructuralModelling, RigidStructuralModelling]
            A model object at the specified index, guaranteed to not be a '$insert'.

        Raises
        ------
        TypeError
            If the value is not one of the concrete types of StructuralModelling; i.e. it is specified with a '$insert'.
        """
        if isinstance(self.Modelling, StructuralModellingInsert) or self.Modelling.is_insert:
            raise TypeError(f"Expected Modelling value to be an in-line object, but it is currently in the '$insert' state.")
        return self.Modelling


    def Modelling_as(self, cls: Type[TStructuralModellingOptions])-> TStructuralModellingOptions:
        """
        Retrieves the value of Modelling, ensuring it is of the specified type.
        The specified type must be one of the valid concrete types of StructuralModelling, or the '$insert' type
        (in the case the model is defined to be inserted from an external resource).

        Useful when using type-checking development tools, as it provides a concise way to access a value with the correct type.

        Parameters
        ----------
        cls: Type[Union[StructuralModellingInsert, ModalStructuralModelling, RigidStructuralModelling]]
            One of the valid concrete types of StructuralModelling, or the '$insert' type
            (in the case the model is defined to be inserted from an external resource).

        Returns
        -------
        TStructuralModellingOptions
            A model object of the specified type.

        Raises
        ------
        TypeError
            If the value is not of the specified type.

        Examples
        --------
        Get a reference to the value when it is one of the types of StructuralModelling:
        >>> val_obj = model_obj.Modelling_as(models.ModalStructuralModelling)
        >>> val_obj = model_obj.Modelling_as(models.RigidStructuralModelling)

        Get a reference to the value, when it was specified with a '$insert' and read in from a file:
        >>> insert_obj = model_obj.Modelling_as(models.StructuralModellingInsert)
        """
        if not isinstance(self.Modelling, cls):
            raise TypeError(f"Expected Modelling of type '{cls.__name__}' but was type '{type(self.Modelling).__name__}'")
        return self.Modelling


FlexBody.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.bladed_model import BladedModel
from dnv_bladed_models.grid_loss_energy_sink import GridLossEnergySink

from .schema_helper import SchemaHelper
from .models_impl import *


class ElectricalGrid(BladedModel):
    r"""
    The definition of the electrical grid that the turbine is connected to.
    
    Not supported yet.
    
    Attributes
    ----------
    NetworkVoltage : float, Not supported yet
        The voltage of the local network.
    
    ConnectingLineResistance : float, Not supported yet
        The resistance of the line that connects the turbine to the local network.
    
    ConnectingLineInductance : float, Not supported yet
        The inductance of the line that connects the turbine to the local network.
    
    NetworkResistance : float, Not supported yet
        The resistance of the network that the turbine is connected to.
    
    NetworkInductance : float, Not supported yet
        The inductance of the network that the turbine is connected to.
    
    NumberOfTurbinesOnFarm : int, Not supported yet
        The number of turbines on the farm that share the same grid connection.
    
    GridLossEnergySinks : List[GridLossEnergySink], Not supported yet
        A list of energy sinks available to the turbine.
    
    Notes
    -----
    This model is not supported yet by the Bladed calculation engine.
    """
    NetworkVoltage: float = Field(alias="NetworkVoltage", default=None) # Not supported yet
    ConnectingLineResistance: float = Field(alias="ConnectingLineResistance", default=None) # Not supported yet
    ConnectingLineInductance: float = Field(alias="ConnectingLineInductance", default=None) # Not supported yet
    NetworkResistance: float = Field(alias="NetworkResistance", default=None) # Not supported yet
    NetworkInductance: float = Field(alias="NetworkInductance", default=None) # Not supported yet
    NumberOfTurbinesOnFarm: int = Field(alias="NumberOfTurbinesOnFarm", default=None) # Not supported yet
    GridLossEnergySinks: List[GridLossEnergySink] = Field(alias="GridLossEnergySinks", default=list()) # Not supported yet

    _relative_schema_path = 'Turbine/ElectricalGrid/ElectricalGrid.json'
    _type_info = TypeInfo(
        set([]),
        set([]),
        set(['GridLossEnergySinks',]),
        None).merge(BladedModel._type_info)


    class Config:
        extra = Extra.forbid
        validate_assignment = True
        allow_population_by_field_name = True
        pass

    def _entity(self) -> bool:
        return True


ElectricalGrid.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.evolving_turbulence import EvolvingTurbulence

from .schema_helper import SchemaHelper
from .models_impl import *


class EvolvingTurbulenceExponential(EvolvingTurbulence):
    r"""
    The settings for Exponential evolving turbulence.  In the case of a normal turbulent wind field, the turbulence is \"frozen\" and approaches the turbine at a constant block.  Although this doesn't match physical reality, it is a particular problem for Lidar - as it gives them a \"perfect\" insight into the oncoming wind field.  In order to represent the nature of real turbulence, a second turbulence file is superimposed on the windfield so that it \"evolves\" as it moves forward.  This is computationally expensive, and is usually applied only to the Lidar readings - although it can be applied to all the wind values in a simulation.
    
    Attributes
    ----------
    EvolvingTurbulenceType : Literal['EvolvingTurbulenceExponential'], default='EvolvingTurbulenceExponential'
        Defines the specific type of EvolvingTurbulence model in use.  For a `EvolvingTurbulenceExponential` object, this must always be set to a value of `EvolvingTurbulenceExponential`.
    
    ExponentialFactor : float
        The exponential factor for use in the exponential model.
    
    Notes
    -----
    
    """
    EvolvingTurbulenceType: Literal['EvolvingTurbulenceExponential'] = Field(alias="EvolvingTurbulenceType", default='EvolvingTurbulenceExponential', allow_mutation=False, const=True) # type: ignore
    ExponentialFactor: float = Field(alias="ExponentialFactor", default=None)

    _relative_schema_path = 'TimeDomainSimulation/Environment/Wind/EvolvingTurbulence/EvolvingTurbulenceExponential.json'
    _type_info = TypeInfo(
        set([]),
        set([]),
        set([]),
        'EvolvingTurbulenceType').merge(EvolvingTurbulence._type_info)


    class Config:
        extra = Extra.forbid
        validate_assignment = True
        allow_population_by_field_name = True
        pass

    def _entity(self) -> bool:
        return True


EvolvingTurbulenceExponential.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.bladed_model import BladedModel

from .schema_helper import SchemaHelper
from .models_impl import *


class PositionOfHubCentre(BladedModel):
    r"""
    The positioning of the hub centre.  The hub centre is the nominal point where all of the pitch axes intercept the axis of rotation, if any sweep is ignored.
    
    Attributes
    ----------
    RotorTilt : float
        The angle of the main bearing and low speed shaft to the horizontal.  A positive value will tilt the rotor upwards to face the sky.
    
    Overhang : float
        The distance of the hub centre upwind from the `DrivetrainAndNacelle`'s origin (where the nacelle is attached to the yaw bearing or support structure). This will be considered as the vector component parallel to the `DrivetrainAndNacelle`'s x-axis if the `DrivetrainAndNacelle`'s axis system is not aligned with global X.
    
    HeightOffset : float
        The distance in the `DrivetrainAndNacelle`'s Z direction between the `DrivetrainAndNacelle`'s origin (where the nacelle is attached to the yaw system or support structure) and the hub centre.
    
    SideOffset : float, default=0
        The distance in the `DrivetrainAndNacelle`'s Y direction between the `DrivetrainAndNacelle`'s origin (where the nacelle is attached to the yaw system or support structure) and the hub centre.  This is often zero or very small.
    
    Notes
    -----
    
    """
    RotorTilt: float = Field(alias="RotorTilt", default=None)
    Overhang: float = Field(alias="Overhang", default=None)
    HeightOffset: float = Field(alias="HeightOffset", default=None)
    SideOffset: float = Field(alias="SideOffset", default=None)

    _relative_schema_path = 'Components/DrivetrainAndNacelle/PositionOfHubCentre/PositionOfHubCentre.json'
    _type_info = TypeInfo(
        set([]),
        set([]),
        set([]),
        None).merge(BladedModel._type_info)


    class Config:
        extra = Extra.forbid
        validate_assignment = True
        allow_population_by_field_name = True
        pass

    def _entity(self) -> bool:
        return True


PositionOfHubCentre.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.transducer_behaviour import TransducerBehaviour
class PositionProportionalIntegralDeriviative_DifferentialGainActionEnum(str, Enum):
    ERROR = "Error"
    FEEDBACK = "Feedback"
    SETPOINT = "Setpoint"

from .schema_helper import SchemaHelper
from .models_impl import *


class PositionProportionalIntegralDeriviative(TransducerBehaviour):
    r"""
    
    
    Not supported yet.
    
    Attributes
    ----------
    TransducerBehaviourType : Literal['PositionProportionalIntegralDeriviative'], default='PositionProportionalIntegralDeriviative', Not supported yet
        Defines the specific type of TransducerBehaviour model in use.  For a `PositionProportionalIntegralDeriviative` object, this must always be set to a value of `PositionProportionalIntegralDeriviative`.
    
    ProportionalGain : float, default=0, Not supported yet
        The gain on the contemporaneous error.
    
    IntegralGain : float, default=0, Not supported yet
        The gain on the integral from time zero till the present of the error signal.
    
    DifferentialGain : float, default=0, Not supported yet
        The gain the filtered derivative of the error. The derivative of the error is passed through a low-pass filter.
    
    DifferentialGainAction : PositionProportionalIntegralDeriviative_DifferentialGainActionEnum, default='Feedback', Not supported yet
        Proportional and integral terms are applied on error.  The derivative term may also be applied on the feedback or setpoint signals.
    
    DifferentialGainTimeConstant : float, default=0, Not supported yet
        The derivative term uses a low-pass filter on input.
    
    DesaturationTimeConstant : float, default=0, Not supported yet
        The time constant for when the output exceeds the limits.
    
    Notes
    -----
    This model is not supported yet by the Bladed calculation engine.
    """
    TransducerBehaviourType: Literal['PositionProportionalIntegralDeriviative'] = Field(alias="TransducerBehaviourType", default='PositionProportionalIntegralDeriviative', allow_mutation=False, const=True) # Not supported yet # type: ignore
    ProportionalGain: float = Field(alias="ProportionalGain", default=None) # Not supported yet
    IntegralGain: float = Field(alias="IntegralGain", default=None) # Not supported yet
    DifferentialGain: float = Field(alias="DifferentialGain", default=None) # Not supported yet
    DifferentialGainAction: PositionProportionalIntegralDeriviative_DifferentialGainActionEnum = Field(alias="DifferentialGainAction", default=None) # Not supported yet
    DifferentialGainTimeConstant: float = Field(alias="DifferentialGainTimeConstant", default=None) # Not supported yet
    DesaturationTimeConstant: float = Field(alias="DesaturationTimeConstant", default=None) # Not supported yet

    _relative_schema_path = 'Turbine/BladedControl/MeasuredSignalProperties/common/PositionProportionalIntegralDeriviative.json'
    _type_info = TypeInfo(
        set([]),
        set([]),
        set([]),
        'TransducerBehaviourType').merge(TransducerBehaviour._type_info)


    class Config:
        extra = Extra.forbid
        validate_assignment = True
        allow_population_by_field_name = True
        pass

    def _entity(self) -> bool:
        return True


PositionProportionalIntegralDeriviative.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.bladed_model import BladedModel
class AdditionalConstrainedWave_MethodEnum(str, Enum):
    OFF = "OFF"
    ON = "ON"
    NON_LINEAR = "NON-LINEAR"

from .schema_helper import SchemaHelper
from .models_impl import *


class AdditionalConstrainedWave(BladedModel):
    r"""
    The definition of additional constrained wave options.
    
    Not supported yet.
    
    Attributes
    ----------
    Method : AdditionalConstrainedWave_MethodEnum, Not supported yet
        The method used to define a constrained wave.
    
    WaveHeight : float, Not supported yet
        The constrained wave height, measured from trough to crest.
    
    Period : float, Not supported yet
        The constrained wave time period when using the Stream Function method.
    
    BlendNonLinearConstrained : bool, default=True, Not supported yet
        If true, the non-linear surface elevation will be bleded in.
    
    TimeOfWaveCrest : float, Not supported yet
        The time of the crest of the constrained wave.
    
    NumberOfConstraints : int, Not supported yet
        The number of constraints.
    
    Notes
    -----
    This model is not supported yet by the Bladed calculation engine.
    """
    Method: AdditionalConstrainedWave_MethodEnum = Field(alias="Method", default=None) # Not supported yet
    WaveHeight: float = Field(alias="WaveHeight", default=None) # Not supported yet
    Period: float = Field(alias="Period", default=None) # Not supported yet
    BlendNonLinearConstrained: bool = Field(alias="BlendNonLinearConstrained", default=None) # Not supported yet
    TimeOfWaveCrest: float = Field(alias="TimeOfWaveCrest", default=None) # Not supported yet
    NumberOfConstraints: int = Field(alias="NumberOfConstraints", default=None) # Not supported yet

    _relative_schema_path = 'TimeDomainSimulation/Environment/SeaState/Waves/AdditionalConstrainedWave/AdditionalConstrainedWave.json'
    _type_info = TypeInfo(
        set([]),
        set([]),
        set([]),
        None).merge(BladedModel._type_info)


    class Config:
        extra = Extra.forbid
        validate_assignment = True
        allow_population_by_field_name = True
        pass

    def _entity(self) -> bool:
        return True


AdditionalConstrainedWave.update_forward_refs()



# *********************************************************************#
# Bladed Python Models API                                             #
# Copyright (c) DNV Services UK Limited (c) 2025. All rights reserved. #
# MIT License (see license file)                                       #
# *********************************************************************#


# coding: utf-8

from __future__ import annotations

from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import os
import re  # noqa: F401
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Type, Union, Callable, Iterable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra,PrivateAttr  # noqa: F401
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.utils import ROOT_KEY
from json import encoder
from dnv_bladed_models.blade import Blade
from dnv_bladed_models.bladed_model import BladedModel
from dnv_bladed_models.component import Component
from dnv_bladed_models.component_insert import ComponentInsert
from dnv_bladed_models.drivetrain_and_nacelle import DrivetrainAndNacelle
from dnv_bladed_models.external_module_component import ExternalModuleComponent
from dnv_bladed_models.fixed_speed_active_damper import FixedSpeedActiveDamper
from dnv_bladed_models.flexibility import Flexibility
from dnv_bladed_models.independent_pitch_hub import IndependentPitchHub
from dnv_bladed_models.lidar import Lidar
from dnv_bladed_models.linear_passive_damper import LinearPassiveDamper
from dnv_bladed_models.pendulum_damper import PendulumDamper
from dnv_bladed_models.pitch_system import PitchSystem
from dnv_bladed_models.rigid_body_point_inertia import RigidBodyPointInertia
from dnv_bladed_models.rigid_body_sixby_six_inertia import RigidBodySixbySixInertia
from dnv_bladed_models.rotation import Rotation
from dnv_bladed_models.superelement import Superelement
from dnv_bladed_models.tower import Tower
from dnv_bladed_models.translation import Translation
from dnv_bladed_models.variable_speed_active_damper import VariableSpeedActiveDamper
from dnv_bladed_models.variable_speed_generator import VariableSpeedGenerator
from dnv_bladed_models.yaw_system import YawSystem

from .schema_helper import SchemaHelper
from .models_impl import *

TComponentOptions = TypeVar('TComponentOptions', Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, ComponentInsert, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem, )

class ComponentLibrary(BladedModel):
    r"""
    A library of Component definitions. Each Component is added with a unique name, used to reference it from nodes in the Turbine Assembly tree.
    
    Attributes
    ----------
    Notes
    -----
    
    """

    _relative_schema_path = 'Turbine/ComponentLibrary/ComponentLibrary.json'
    _type_info = TypeInfo(
        set([]),
        set([]),
        set([]),
        None).merge(BladedModel._type_info)


    class Config:
        extra = Extra.allow
        validate_assignment = True
        allow_population_by_field_name = True
        pass

    def items_as_inline(self) -> Iterable[Tuple[str, Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]]]:
        """
        Returns an iterable of key-value pairs for all of the user-supplied entries currently in the library.
        If any of the values are specified with a '$insert', an error is raised.

        Useful when using type-checking development tools, and the type of every entry in the library is known to not be an insert.

        Returns
        -------
        Iterable[Tuple[str, Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]]]
            A list of model objects, guaranteed to not be an 'insert'.

        Raises
        ------
        TypeError
            If any of the values are not one of the concrete types of Component; i.e. they are specified with a '$insert'.

        Examples
        --------
        >>> for key, Component_obj in obj.items_as_inline():
                print(f"Entry '{key}' is of type Component_obj.ComponentType}")
        """
        for key, val in self.items():
            if isinstance(val, ComponentInsert) or val.is_insert == True:
                raise TypeError(f"Expected element value to be an in-line object, but it is currently in the '$insert' state.")
            yield (key, val)


    def items_as(self, entry_cls: Type[TComponentOptions]) -> Iterable[Tuple[str, TComponentOptions]]:
        """
        Returns an iterable of key-value pairs for all of the user-supplied entries currently in the library,
        ensuring each value is of the specified type, otherwise an error is raised.

        Useful when using type-checking development tools, and the type of every entry in the library is known to be of the same type.

        Returns
        -------
        Iterable[Tuple[str, TComponentOptions]
            A list of key-value tuples, with each value guaranteed to be of the specified type.

        Raises
        ------
        TypeError
            If any of the values are not one of the specified type.

        Examples
        --------
        Get a reference to each entry in the library when each one is a type of Component:
        >>> for key, Blade_obj in obj.items_as(models.Blade):
                # process object
        >>> for key, DrivetrainAndNacelle_obj in obj.items_as(models.DrivetrainAndNacelle):
                # process object
        >>> for key, ExternalModuleComponent_obj in obj.items_as(models.ExternalModuleComponent):
                # process object
        >>> for key, FixedSpeedActiveDamper_obj in obj.items_as(models.FixedSpeedActiveDamper):
                # process object
        >>> for key, Flexibility_obj in obj.items_as(models.Flexibility):
                # process object
        >>> for key, IndependentPitchHub_obj in obj.items_as(models.IndependentPitchHub):
                # process object
        >>> for key, Lidar_obj in obj.items_as(models.Lidar):
                # process object
        >>> for key, LinearPassiveDamper_obj in obj.items_as(models.LinearPassiveDamper):
                # process object
        >>> for key, PendulumDamper_obj in obj.items_as(models.PendulumDamper):
                # process object
        >>> for key, PitchSystem_obj in obj.items_as(models.PitchSystem):
                # process object
        >>> for key, RigidBodyPointInertia_obj in obj.items_as(models.RigidBodyPointInertia):
                # process object
        >>> for key, RigidBodySixbySixInertia_obj in obj.items_as(models.RigidBodySixbySixInertia):
                # process object
        >>> for key, Rotation_obj in obj.items_as(models.Rotation):
                # process object
        >>> for key, Superelement_obj in obj.items_as(models.Superelement):
                # process object
        >>> for key, Tower_obj in obj.items_as(models.Tower):
                # process object
        >>> for key, Translation_obj in obj.items_as(models.Translation):
                # process object
        >>> for key, VariableSpeedActiveDamper_obj in obj.items_as(models.VariableSpeedActiveDamper):
                # process object
        >>> for key, VariableSpeedGenerator_obj in obj.items_as(models.VariableSpeedGenerator):
                # process object
        >>> for key, YawSystem_obj in obj.items_as(models.YawSystem):
                # process object

        Get a reference to each entry in the library, when each one was specified with a '$insert' and read in from a file:
        >>> for key, insert_obj in obj.items_as(models.ComponentInsert):
                # process object
        """
        for key, val in self.items():
            if not isinstance(val, entry_cls):
                raise ValueError(f"Expected value of type '{entry_cls.__name__}' for key '{key}' but found type '{type(val).__name__}'")
            yield (key, val)


    def Component_as_Blade(self, key: Union[str, int]) -> Blade:
        """
        Retrieves a Component entry from the library, guaranteeing it is a Blade; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        Blade
            A model object guaranteed to be a Blade.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a Blade.

        Examples
        --------
        >>> Blade_obj = library_obj.Component_as_Blade('my_entry_key')
        """
        return self.Component_as(key, Blade)


    def Component_as_DrivetrainAndNacelle(self, key: Union[str, int]) -> DrivetrainAndNacelle:
        """
        Retrieves a Component entry from the library, guaranteeing it is a DrivetrainAndNacelle; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        DrivetrainAndNacelle
            A model object guaranteed to be a DrivetrainAndNacelle.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a DrivetrainAndNacelle.

        Examples
        --------
        >>> DrivetrainAndNacelle_obj = library_obj.Component_as_DrivetrainAndNacelle('my_entry_key')
        """
        return self.Component_as(key, DrivetrainAndNacelle)


    def Component_as_ExternalModuleComponent(self, key: Union[str, int]) -> ExternalModuleComponent:
        """
        Retrieves a Component entry from the library, guaranteeing it is a ExternalModuleComponent; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        ExternalModuleComponent
            A model object guaranteed to be a ExternalModuleComponent.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a ExternalModuleComponent.

        Examples
        --------
        >>> ExternalModuleComponent_obj = library_obj.Component_as_ExternalModuleComponent('my_entry_key')
        """
        return self.Component_as(key, ExternalModuleComponent)


    def Component_as_FixedSpeedActiveDamper(self, key: Union[str, int]) -> FixedSpeedActiveDamper:
        """
        Retrieves a Component entry from the library, guaranteeing it is a FixedSpeedActiveDamper; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        FixedSpeedActiveDamper
            A model object guaranteed to be a FixedSpeedActiveDamper.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a FixedSpeedActiveDamper.

        Examples
        --------
        >>> FixedSpeedActiveDamper_obj = library_obj.Component_as_FixedSpeedActiveDamper('my_entry_key')
        """
        return self.Component_as(key, FixedSpeedActiveDamper)


    def Component_as_Flexibility(self, key: Union[str, int]) -> Flexibility:
        """
        Retrieves a Component entry from the library, guaranteeing it is a Flexibility; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        Flexibility
            A model object guaranteed to be a Flexibility.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a Flexibility.

        Examples
        --------
        >>> Flexibility_obj = library_obj.Component_as_Flexibility('my_entry_key')
        """
        return self.Component_as(key, Flexibility)


    def Component_as_IndependentPitchHub(self, key: Union[str, int]) -> IndependentPitchHub:
        """
        Retrieves a Component entry from the library, guaranteeing it is a IndependentPitchHub; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        IndependentPitchHub
            A model object guaranteed to be a IndependentPitchHub.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a IndependentPitchHub.

        Examples
        --------
        >>> IndependentPitchHub_obj = library_obj.Component_as_IndependentPitchHub('my_entry_key')
        """
        return self.Component_as(key, IndependentPitchHub)


    def Component_as_Lidar(self, key: Union[str, int]) -> Lidar:
        """
        Retrieves a Component entry from the library, guaranteeing it is a Lidar; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        Lidar
            A model object guaranteed to be a Lidar.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a Lidar.

        Examples
        --------
        >>> Lidar_obj = library_obj.Component_as_Lidar('my_entry_key')
        """
        return self.Component_as(key, Lidar)


    def Component_as_LinearPassiveDamper(self, key: Union[str, int]) -> LinearPassiveDamper:
        """
        Retrieves a Component entry from the library, guaranteeing it is a LinearPassiveDamper; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        LinearPassiveDamper
            A model object guaranteed to be a LinearPassiveDamper.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a LinearPassiveDamper.

        Examples
        --------
        >>> LinearPassiveDamper_obj = library_obj.Component_as_LinearPassiveDamper('my_entry_key')
        """
        return self.Component_as(key, LinearPassiveDamper)


    def Component_as_PendulumDamper(self, key: Union[str, int]) -> PendulumDamper:
        """
        Retrieves a Component entry from the library, guaranteeing it is a PendulumDamper; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        PendulumDamper
            A model object guaranteed to be a PendulumDamper.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a PendulumDamper.

        Examples
        --------
        >>> PendulumDamper_obj = library_obj.Component_as_PendulumDamper('my_entry_key')
        """
        return self.Component_as(key, PendulumDamper)


    def Component_as_PitchSystem(self, key: Union[str, int]) -> PitchSystem:
        """
        Retrieves a Component entry from the library, guaranteeing it is a PitchSystem; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        PitchSystem
            A model object guaranteed to be a PitchSystem.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a PitchSystem.

        Examples
        --------
        >>> PitchSystem_obj = library_obj.Component_as_PitchSystem('my_entry_key')
        """
        return self.Component_as(key, PitchSystem)


    def Component_as_RigidBodyPointInertia(self, key: Union[str, int]) -> RigidBodyPointInertia:
        """
        Retrieves a Component entry from the library, guaranteeing it is a RigidBodyPointInertia; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        RigidBodyPointInertia
            A model object guaranteed to be a RigidBodyPointInertia.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a RigidBodyPointInertia.

        Examples
        --------
        >>> RigidBodyPointInertia_obj = library_obj.Component_as_RigidBodyPointInertia('my_entry_key')
        """
        return self.Component_as(key, RigidBodyPointInertia)


    def Component_as_RigidBodySixbySixInertia(self, key: Union[str, int]) -> RigidBodySixbySixInertia:
        """
        Retrieves a Component entry from the library, guaranteeing it is a RigidBodySixbySixInertia; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        RigidBodySixbySixInertia
            A model object guaranteed to be a RigidBodySixbySixInertia.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a RigidBodySixbySixInertia.

        Examples
        --------
        >>> RigidBodySixbySixInertia_obj = library_obj.Component_as_RigidBodySixbySixInertia('my_entry_key')
        """
        return self.Component_as(key, RigidBodySixbySixInertia)


    def Component_as_Rotation(self, key: Union[str, int]) -> Rotation:
        """
        Retrieves a Component entry from the library, guaranteeing it is a Rotation; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        Rotation
            A model object guaranteed to be a Rotation.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a Rotation.

        Examples
        --------
        >>> Rotation_obj = library_obj.Component_as_Rotation('my_entry_key')
        """
        return self.Component_as(key, Rotation)


    def Component_as_Superelement(self, key: Union[str, int]) -> Superelement:
        """
        Retrieves a Component entry from the library, guaranteeing it is a Superelement; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        Superelement
            A model object guaranteed to be a Superelement.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a Superelement.

        Examples
        --------
        >>> Superelement_obj = library_obj.Component_as_Superelement('my_entry_key')
        """
        return self.Component_as(key, Superelement)


    def Component_as_Tower(self, key: Union[str, int]) -> Tower:
        """
        Retrieves a Component entry from the library, guaranteeing it is a Tower; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        Tower
            A model object guaranteed to be a Tower.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a Tower.

        Examples
        --------
        >>> Tower_obj = library_obj.Component_as_Tower('my_entry_key')
        """
        return self.Component_as(key, Tower)


    def Component_as_Translation(self, key: Union[str, int]) -> Translation:
        """
        Retrieves a Component entry from the library, guaranteeing it is a Translation; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        Translation
            A model object guaranteed to be a Translation.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a Translation.

        Examples
        --------
        >>> Translation_obj = library_obj.Component_as_Translation('my_entry_key')
        """
        return self.Component_as(key, Translation)


    def Component_as_VariableSpeedActiveDamper(self, key: Union[str, int]) -> VariableSpeedActiveDamper:
        """
        Retrieves a Component entry from the library, guaranteeing it is a VariableSpeedActiveDamper; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        VariableSpeedActiveDamper
            A model object guaranteed to be a VariableSpeedActiveDamper.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a VariableSpeedActiveDamper.

        Examples
        --------
        >>> VariableSpeedActiveDamper_obj = library_obj.Component_as_VariableSpeedActiveDamper('my_entry_key')
        """
        return self.Component_as(key, VariableSpeedActiveDamper)


    def Component_as_VariableSpeedGenerator(self, key: Union[str, int]) -> VariableSpeedGenerator:
        """
        Retrieves a Component entry from the library, guaranteeing it is a VariableSpeedGenerator; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        VariableSpeedGenerator
            A model object guaranteed to be a VariableSpeedGenerator.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a VariableSpeedGenerator.

        Examples
        --------
        >>> VariableSpeedGenerator_obj = library_obj.Component_as_VariableSpeedGenerator('my_entry_key')
        """
        return self.Component_as(key, VariableSpeedGenerator)


    def Component_as_YawSystem(self, key: Union[str, int]) -> YawSystem:
        """
        Retrieves a Component entry from the library, guaranteeing it is a YawSystem; if it is not, an error is raised.

        Useful when using type-checking development tools, and an entry is expected to be a specific type, and not specified with a '$insert'.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        YawSystem
            A model object guaranteed to be a YawSystem.

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not a YawSystem.

        Examples
        --------
        >>> YawSystem_obj = library_obj.Component_as_YawSystem('my_entry_key')
        """
        return self.Component_as(key, YawSystem)


    def Component_as_inline(self, key: Union[str, int]) -> Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]:
        """
        Retrieves a Component from the library, if the value is specified with a '$insert', an error is raised.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        Returns
        -------
        Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]
            A model object that could be any of the concrete types of Component specified in-line (and not a '$insert').

        Raises
        ------
        KeyError
            If the key does not exist in the library.

        TypeError
            If the value is not one of the concrete types of Component; i.e. it is specified with a '$insert'.

        Examples
        --------
        >>> Component_obj = library_obj.Component_as_inline('my_entry_key')
        """
        val = self.__getitem__(key)
        if isinstance(val, ComponentInsert) or val.is_insert == True:
            raise TypeError(f"Expected entry value at '{key}' to be an in-line object, but it is currently in the 'insert' state.")
        return val


    def Component_as(self, key: Union[str, int], entry_cls: Type[TComponentOptions]) -> TComponentOptions:
        """
        Retrieves an object from the library, ensuring it is of the specified type.
        The specified type must be one of the valid concrete types of Component, or the '$insert' type
        (in the case the model is defined to be inserted from an external resource).

        Useful when using type-checking development tools, as it provides a concise way to access a value with the correct type.

        Parameters
        ----------
        key: Union[str, int]
            The key or index of the entry to retrieve.

        entry_cls: Type[Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, ComponentInsert, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]]
            One of the valid concrete types of Component, or the '$insert' type
            (in the case the model is defined to be inserted from an external resource).

        Returns
        -------
        TComponentOptions
            A model object of the specified type.

        Raises
        ------
        KeyError
            If the key does not exist in the library, or if the index is out of range.

        ValueError
            If the model object for the specified key is not of the specified type.

        Examples
        --------
        Get a reference to an entry as one of the types of Component (including the '$insert' type):
        >>> Blade_obj = library_obj.Component_as('my_entry_key', models.Blade)
        >>> DrivetrainAndNacelle_obj = library_obj.Component_as('my_entry_key', models.DrivetrainAndNacelle)
        >>> ExternalModuleComponent_obj = library_obj.Component_as('my_entry_key', models.ExternalModuleComponent)
        >>> FixedSpeedActiveDamper_obj = library_obj.Component_as('my_entry_key', models.FixedSpeedActiveDamper)
        >>> Flexibility_obj = library_obj.Component_as('my_entry_key', models.Flexibility)
        >>> IndependentPitchHub_obj = library_obj.Component_as('my_entry_key', models.IndependentPitchHub)
        >>> Lidar_obj = library_obj.Component_as('my_entry_key', models.Lidar)
        >>> LinearPassiveDamper_obj = library_obj.Component_as('my_entry_key', models.LinearPassiveDamper)
        >>> PendulumDamper_obj = library_obj.Component_as('my_entry_key', models.PendulumDamper)
        >>> PitchSystem_obj = library_obj.Component_as('my_entry_key', models.PitchSystem)
        >>> RigidBodyPointInertia_obj = library_obj.Component_as('my_entry_key', models.RigidBodyPointInertia)
        >>> RigidBodySixbySixInertia_obj = library_obj.Component_as('my_entry_key', models.RigidBodySixbySixInertia)
        >>> Rotation_obj = library_obj.Component_as('my_entry_key', models.Rotation)
        >>> Superelement_obj = library_obj.Component_as('my_entry_key', models.Superelement)
        >>> Tower_obj = library_obj.Component_as('my_entry_key', models.Tower)
        >>> Translation_obj = library_obj.Component_as('my_entry_key', models.Translation)
        >>> VariableSpeedActiveDamper_obj = library_obj.Component_as('my_entry_key', models.VariableSpeedActiveDamper)
        >>> VariableSpeedGenerator_obj = library_obj.Component_as('my_entry_key', models.VariableSpeedGenerator)
        >>> YawSystem_obj = library_obj.Component_as('my_entry_key', models.YawSystem)

        Get a reference to the value, when it was specified with a '$insert' and read in from a file:
        >>> insert_obj = library_obj.Component_as('my_entry_key', models.ComponentInsert)
        """
        val = self.__getitem__(key)
        if not isinstance(val, entry_cls):
            raise ValueError(f"Expected value of type '{entry_cls.__name__}' for key '{key}' but found type '{type(val).__name__}'")
        return val


    def items(self) -> list[tuple[str, Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, ComponentInsert, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]]]:
        """
        Returns a list of key-value pairs for all of the user-supplied entries currently in the model.
        """
        return [(k, self.__dict__[k]) for k in self.__dict__ if k not in self.__fields__]


    def keys(self) -> list[str]:
        """
        Returns a list of keys for all of the user-supplied entries currently in the model.
        """
        return [k for k in self.__dict__ if k not in self.__fields__]


    def values(self) -> list[Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, ComponentInsert, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]]:
        """
        Returns a list of model objects for all of the user-supplied entries currently in the model.
        """
        return [self.__dict__[k] for k in self.__dict__ if k not in self.__fields__]


    def __len__(self):
        return len([k for k in self.__dict__ if k not in self.__fields__])


    def __contains__(self, item):
        for k in self.__dict__:
            if k not in self.__fields__ and k == item:
                return True
        return False


    def __getitem__(self, key: Union[str, int]) -> Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, ComponentInsert, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]:
        if isinstance(key, int):
            keys = self.keys()
            if len(keys) == 0:
                raise KeyError(f"There are currently no entries in the model object.")
            if key < 0 or key >= len(keys):
                raise KeyError(f"Invalid index specified: {key} (0 >= i < {len(keys)})")
            key = keys[key]
        elif isinstance(key, str):
            if key in self.__fields__:
                raise KeyError(f"Item accessors can only be used for custom entries. '{key}' is a pre-defined model property.")
            if not key in self.__dict__:
                raise KeyError(f"There is no entry with key '{key}'.")
        return getattr(self, key)


    def __setitem__(self, key: str, value: Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, ComponentInsert, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]):
        if not isinstance(key, str):
            raise KeyError(f"Custom entries can only be added with string keys")
        if key in self.__fields__:
            raise KeyError(f"Item accessors can only be used for custom entries. '{key}' is a pre-defined model property.")
        if not isinstance(value, Component) and not isinstance(value, ComponentInsert):
            raise TypeError(f"Entries must be of type 'Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, ComponentInsert, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]'; received '{type(value).__name__}'")
        setattr(self, key, value)


    def __delitem__(self, key: Union[str, int]):
        if isinstance(key, int):
            keys = self.keys()
            if len(keys) == 0:
                raise KeyError(f"There are currently no entries in the model object.")
            if key < 0 or key >= len(keys):
                raise KeyError(f"Invalid index specified: {key} (0 >= i < {len(keys)})")
            key = keys[key]
        elif isinstance(key, str):
            if key in self.__fields__:
                raise KeyError(f"Item accessors can only be used for custom entries. '{key}' is a pre-defined model property.")
            if not key in self.__dict__:
                raise KeyError(f"There is no entry with key '{key}'.")
        delattr(self, key)


    def __setattr__(self, name: str, value: Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, ComponentInsert, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]):
        if not name in self.__fields__ and not isinstance(value, Component) and not isinstance(value, ComponentInsert):
            raise TypeError(f"Entries must be of type 'Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, ComponentInsert, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem]'; received '{type(value).__name__}'")
        super().__setattr__(name, value)



    class __CustomEntries__(BaseModel):
        entries: Dict[str, Annotated[Union[Blade, DrivetrainAndNacelle, ExternalModuleComponent, FixedSpeedActiveDamper, Flexibility, IndependentPitchHub, ComponentInsert, Lidar, LinearPassiveDamper, PendulumDamper, PitchSystem, RigidBodyPointInertia, RigidBodySixbySixInertia, Rotation, Superelement, Tower, Translation, VariableSpeedActiveDamper, VariableSpeedGenerator, YawSystem], Field(discriminator='ComponentType')]]


    @classmethod
    def _model_factory(cls: Type['Model'], obj: dict[str, Any]) -> Model:
        return custom_entries_parser(cls, obj, dict, ComponentLibrary.__CustomEntries__, lambda entry_key, entry_val: prepare_dict_for_discriminated_insert(cls, entry_key, entry_val, 'ComponentType'))

    def _entity(self) -> bool:
        return True


ComponentLibrary.update_forward_refs()
ComponentLibrary.__CustomEntries__.update_forward_refs()
